C$Procedure  RDFFDI   ( read and parse flatfile records )
 
      SUBROUTINE RDFFDI ( INFILE, NREC, FORMAT, ND, NI, ARD, ARI, EOF )
 
C$ Abstract
C
C     Read a flatfile record and parse it into integer and/or double
C     precision values according to a given format specification.
C
C$ Copyright
C
C     Copyright (1997), California Institute of Technology.
C     U.S. Government sponsorship acknowledged.
C
C$ Required_Reading
C
C     None.
C
C$ Keywords
C
C     FLATFILE
C
C$ Declarations
 
 
      IMPLICIT NONE
 
 
C     Maximum allowed length of a single field.
 
      INTEGER               MAXWD
      PARAMETER           ( MAXWD  = 40 )
 
 
C     Maximum length of an input data record.
 
 
      INTEGER               MAXSTR
      PARAMETER           ( MAXSTR = 80 )
 
 
C     Maximum number of items in format and input data records.
 
 
      INTEGER               MAXITM
      PARAMETER           ( MAXITM = 6 )
 
      LOGICAL               EOF
 
      INTEGER               ND
      INTEGER               NI
      INTEGER               NREC
      INTEGER               ARI     ( * )
 
      CHARACTER*(*)         FORMAT
      CHARACTER*(*)         INFILE
 
      DOUBLE PRECISION      ARD     ( * )
 
 
C$ Brief_I/O
C
C     VARIABLE  I/O  DESCRIPTION
C     --------  ---  --------------------------------------------------
C     INFILE     I   Name of input flatfile.
C     NREC       I   Current record number in flatfile.
C     FORMAT     I   String containing format picture.
C     ND         O   Number of double precision values found.
C     NI         O   Number of integer values found.
C     ARD        O   Array of double precision values read.
C     ARI        O   Array of integer values read.
C     EOF        O   End of file flag.
C
C
C$ Detailed_Input
C
C     INFILE     is the full pathname of the input flatfile.
C
C     NREC       is the current record number in the flatfile,
C                provided by the user.
C
C     FORMAT     is a string containing a format description or
C                picture of the data to be parsed.  In the present
C                scheme, 'I' signifies integer and 'D' double
C                precision values.  A typical format string might
C                be 'I D D D'.  Up to MAXITM format items are permitted.
C
C$ Detailed_Output
C
C     ND         is the number of double precision values found.
C
C     NI         is the number of integer values found.
C
C     ARD        is an array containing consecutively read
C                double precision values.
C
C     ARI        is an array containing consecutively read
C                integer values.
C
C     EOF        is .TRUE. if end of file was encountered.
C
C$ Parameters
C
C     None.
C
C$ Files
C
C     None.
C
C$ Exceptions
C
C     1)  If the format string length exceeds MAXSTR, then
C         SPICE(FORMATSTRINGTOOLONG) is signaled.
C
C     2)  If the number of format items exceeds MAXITM, then
C         SPICE(FORMATITEMLIMITEXCEEDED) is signaled.
C
C     3)  If the number of data items exceeds MAXITM, then
C         SPICE(DATAITEMLIMITEXCEEDED) is signaled.
C
C     4)  If an unknown format type occurs in the format string,
C         then SPICE(BADFORMATSPECIFIER) is signaled.
C
C     5)  If a string representing numerical data is exactly
C         MAXWD in length then SPICE(DATAWIDTHERROR) is signaled.
C
C     6)  If the format string and data string contain a
C         different number of items delimited by spaces,
C         then SPICE(FORMATDATAMISMATCH) is signaled.
C
C     7)  If an error is returned from NPARSI, then SPICE(BADINTEGER)
C         is signaled.
C
C     8)  If an error is returned from NPARSD, then
C         SPICE(BADDOUBLEPRECISION) is signaled.
C
C
C$ Particulars
C
C     1)  It is the responsibility of the programmer to supply the
C         correct record number in the file, NREC.
C
C     2)  Format and data items are delimited by spaces.
C
C$ Examples
C
C     C
C     C     Read a "self-describing" flatfile:
C     C
C     C     The first record contains a single integer,
C     C     which is the number of data records to be read.
C     C
C     C     The following line contains a format string for the
C     C     data records to follow.
C     C
C     C     After the last data record is read, a new data group
C     C     may begin, signified by a line containing an
C     C     integer count, followed by a format string and
C     C     data records.  Alternatively, the end of the file may
C     C     also be encountered.
C     C
C
C
C           LOGICAL               EOF
C           CHARACTER*(*)         FNAME
C           CHARACTER*(MAXSTR)    FORMAT
C           INTEGER               ND
C           INTEGER               NI
C           INTEGER               NREC
C           INTEGER               NDAT
C           INTEGER               ARI     ( MAXITM )
C           DOUBLE PRECISION      ARD     ( MAXITM )
C
C
C           CALL PROMPT ( 'Enter file name ', FNAME )
C
C     C
C     C     Get record count.
C     C
C
C           NREC = 1
C           CALL RDFFDI ( FNAME, NREC, 'I', ND, NI, ARD, ARI, EOF )
C
C           DO WHILE ( .NOT. EOF )
C
C              NDAT = ARI(1)
C
C     C
C     C        Get data record format.
C     C
C
C              NREC = NREC + 1
C              CALL RDTEXT ( FNAME, FORMAT, EOF )
C
C              IF ( EOF ) THEN
C                 CALL EXTMSI ( 'End of file after line #.', '#',
C           .                    NREC-1 )
C              END IF
C
C     C
C     C        Read data records.
C     C
C
C              DO I = 1, NDAT
C                 NREC = NREC + 1
C                 CALL RDFFDI ( FNAME, NREC, FORMAT, ND, NI,
C           .                   ARD,   ARI,  EOF )
C                 IF ( EOF ) THEN
C                    CALL EXTMSI ( 'End of file after line #.', '#',
C           .                       NREC-1 )
C                 END IF
C                 CALL HAVFUN ( ND, NI, ARD, ARI )
C              END DO
C
C     C
C     C        Get next data record count, if it's there.
C     C
C
C              NREC = NREC + 1
C              CALL RDFFDI ( FNAME, NREC, 'I', ND, NI, ARD, ARI, EOF )
C
C           END DO
C
C
C$ Restrictions
C
C     None.
C
C$ Author_and_Institution
C
C     J.A. Bytof      (JPL)
C
C$ Literature_References
C
C     None.
C
C$ Version
C
C-    SPICELIB Version 1.0.0, 09-APRIL-1997 (JAB)
C
C-&
 
C$ Index_Entries
C
C     read and parse flatfile records.
C
C-&
 
C
C     SPICELIB functions
C
      INTEGER               WDCNT
 
C
C     Local variables.
C
 
      INTEGER               I
      INTEGER               PTR
      INTEGER               NFMT
      INTEGER               NDAT
      INTEGER               DATI
      INTEGER               LASTNB
 
      LOGICAL               RETURN
 
      CHARACTER*(160)       ERROR
      CHARACTER*(MAXSTR)    LINE
      CHARACTER*(MAXSTR)    FMTSAV
      CHARACTER*(MAXSTR)    FMTITM     ( MAXITM )
      CHARACTER*(MAXSTR)    DATITM     ( MAXITM )
 
      DOUBLE PRECISION      DATD
 
      SAVE                  NFMT
      SAVE                  FMTITM
      SAVE                  FMTSAV
 
      DATA                  FMTSAV     / 'X' /
 
 
C
C     Standard SPICE error handling.
C
      IF ( RETURN () ) THEN
         RETURN
      ELSE
         CALL CHKIN ( 'RDFFDI' )
      END IF
 
C
C     Initialize array counters and logical flags.
C
      ND     = 0
      NI     = 0
 
      EOF    = .FALSE.
 
C
C     Check format string length.
C
 
      IF ( LASTNB(FORMAT) .GT. MAXSTR ) THEN
         CALL SETMSG (     'Format string length exceeds limit'
     .                 //  ' in line #.'                          )
         CALL ERRINT ( '#', NREC                                  )
         CALL SIGERR ( 'SPICE(FORMATSTRINGTOOLONG)'               )
      END IF
 
C
C     Parse format string if it has changed.  Verify that format.
C
      IF ( FORMAT .NE. FMTSAV ) THEN
 
C
C     Does the number of format items exceed MAXITM?
C
         IF ( WDCNT( FORMAT ) .GT. MAXITM ) THEN
            CALL SETMSG ( 'Too many format items at line #'      )
            CALL ERRINT ( '#', NREC                              )
            CALL SIGERR ( 'SPICE(FMTITEMLIMITEXCEEDED)'          )
         END IF
 
         CALL LPARSE ( FORMAT,  ' ', MAXITM, NFMT, FMTITM )
 
         DO I = 1, NFMT
            IF ( LASTNB(FMTITM(I)) .GT. 1 ) THEN
               CALL SETMSG ( 'Format error detected at line #.'  )
               CALL ERRINT ( '#', NREC                           )
               CALL SIGERR ( 'SPICE(BADFORMATSPECIFIER)'         )
            END IF
         END DO
         FMTSAV = FORMAT
 
      END IF
 
 
C
C     Read data record, if end of file is encountered,
C     return immediately to calling routine.
C
      CALL RDTEXT ( INFILE, LINE, EOF )
 
      IF ( EOF ) RETURN
 
C
C     Convert any tabs or carriage returns to blanks.
C
      CALL  REPLCH ( LINE, CHAR( 9), ' ', LINE )
      CALL  REPLCH ( LINE, CHAR(13), ' ', LINE )
 
C
C     Does the number of data items exceed MAXITM?
C
      NDAT = WDCNT ( LINE )
 
      IF ( NDAT .GT. MAXITM ) THEN
         CALL SETMSG ( 'Too many data items, line #'                 )
         CALL ERRINT ( '#', NREC                                     )
         CALL SIGERR ( 'SPICE(DATAITEMLIMITEXCEEDED)'                )
      END IF
 
C
C     The format specifier and data record might mistakenly
C     have a different number of fields.
C
      IF (      NFMT .LT. NDAT ) THEN
 
         CALL SETMSG ( 'Too many data items in line #.'    )
         CALL ERRINT ( '#', NREC                           )
         CALL SIGERR ( 'SPICE(FORMATDATAMISMATCH)'         )
 
      ELSE IF ( NFMT .GT. NDAT ) THEN
 
         CALL SETMSG ( 'Too many format items in line #.'  )
         CALL ERRINT ( '#', NREC                           )
         CALL SIGERR ( 'SPICE(FORMATDATAMISMATCH)'         )
 
      END IF
 
C
C     Parse data, verify that values fit within MAXWD field width.
C
      CALL LPARSE ( LINE,  ' ', MAXITM, NDAT, DATITM )
      DO I = 1, NDAT
         IF ( LASTNB(DATITM(I)) .GE. MAXWD ) THEN
            CALL SETMSG ( 'Possible data error in line #.'    )
            CALL ERRINT ( '#', NREC                           )
            CALL SIGERR ( 'SPICE(DATAWIDTHERROR)'             )
         END IF
      END DO
 
C
C     Process each data field according to its format.
C
      DO I = 1, NDAT
 
C
C     Convert the data field to requested type.
C
         IF ( FMTITM(I)(1:1) .EQ. 'I' ) THEN
 
            CALL NPARSI ( DATITM(I), DATI, ERROR, PTR )
            IF ( PTR .NE. 0 ) THEN
               CALL SETMSG ( 'Integer error (#) in line #.'   )
               CALL ERRCH  ( '#', ERROR                       )
               CALL ERRINT ( '#', NREC                        )
               CALL SIGERR ( 'SPICE(BADINTEGER)'              )
            END IF
 
            NI = NI + 1
            ARI ( NI ) = DATI
 
         ELSE IF (FMTITM(I)(1:1) .EQ. 'D' ) THEN
 
            CALL NPARSD ( DATITM(I), DATD, ERROR, PTR )
            IF ( PTR .NE. 0 ) THEN
               CALL SETMSG ( 'D.P. error (#) in line #.'     )
               CALL ERRCH  ( '#', ERROR                      )
               CALL ERRINT ( '#', NREC                       )
               CALL SIGERR ( 'SPICE(BADDOUBLEPRECISION)'     )
            END IF
 
            ND = ND + 1
            ARD ( ND ) = DATD
 
C
C        Problem with format specifier.
C
         ELSE
 
            CALL SETMSG ( 'Bad format specifier # used in line #.' )
            CALL ERRCH  ( '#', FMTITM(I)                           )
            CALL ERRINT ( '#', NREC                                )
            CALL SIGERR ( 'SPICE(BADFORMATSPECIFIER)'              )
 
         END IF
 
      END DO
 
      CALL CHKOUT ( 'RDFFDI' )
 
      RETURN
      END
